#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2008 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;
using HeuristicLab.PluginInfrastructure;
using System.Net;
using System.ServiceModel;
using HeuristicLab.CEDMA.DB.Interfaces;
using HeuristicLab.CEDMA.DB;
using System.ServiceModel.Description;
using System.Linq;
using HeuristicLab.CEDMA.Core;
using HeuristicLab.GP.StructureIdentification;
using HeuristicLab.Data;
using HeuristicLab.Core;

namespace HeuristicLab.CEDMA.Server {
  public class Dispatcher {
    private List<Execution> dispatchQueue;
    public IList<string> DispatchQueue {
      get { return dispatchQueue.Select(t => "StandardGP").ToList(); }
    }

    private IStore store;

    public Dispatcher(IStore store) {
      this.store = store;
      this.dispatchQueue = new List<Execution>();
    }

    private void FillDispatchQueue() {
      Dictionary<Entity, Dictionary<int, int>> numberOfModelsOfTargetVariableOfDataSet = new Dictionary<Entity, Dictionary<int, int>>();
      var datasetStatements = store.Select(new Statement(Ontology.AnyEntity, Ontology.PredicateInstanceOf, Ontology.TypeDataSet));
      foreach (Statement datasetStatement in datasetStatements) {
        numberOfModelsOfTargetVariableOfDataSet.Add(datasetStatement.Subject, new Dictionary<int, int>());
        var modelStatements = store.Select(new Statement(datasetStatement.Subject, Ontology.PredicateHasModel, Ontology.AnyEntity));
        foreach (Statement modelStatement in modelStatements) {
          var modelAttributeStatements = store.Select(new Statement((Entity)modelStatement.Property, Ontology.PredicateModelAttribute, Ontology.AnyEntity));
          foreach (Statement modelAttrStatement in modelAttributeStatements) {
            var targetVariableStatements = store.Select(new Statement((Entity)modelAttrStatement.Property, Ontology.PredicateModelAttributeName, Ontology.AnyEntity))
               .Where(t => (string)((Literal)t.Property).Value == "TargetVariable")
               .SelectMany(t => store.Select(new Statement((Entity)modelAttrStatement.Property, Ontology.PredicateModelAttributeValue, Ontology.AnyEntity)))
               .GroupBy(t => (int)((Literal)t.Property).Value);
            foreach (var targetVariable in targetVariableStatements) {
              if (numberOfModelsOfTargetVariableOfDataSet[datasetStatement.Subject].ContainsKey(targetVariable.Key)) {
                numberOfModelsOfTargetVariableOfDataSet[datasetStatement.Subject][targetVariable.Key] = targetVariable.Count();
              } else {
                numberOfModelsOfTargetVariableOfDataSet[datasetStatement.Subject].Add(targetVariable.Key, targetVariable.Count());
              }
            }
          }
        }
      }
      foreach (KeyValuePair<Entity, Dictionary<int, int>> dataSetEntry in numberOfModelsOfTargetVariableOfDataSet) {
        DataSet dataSet = new DataSet(store, dataSetEntry.Key);
        foreach (int targetVariable in dataSet.Problem.AllowedTargetVariables)
          if (!dataSetEntry.Value.ContainsKey(targetVariable) || dataSetEntry.Value[targetVariable] < 10) {
            IEngine engine = CreateEngine(dataSet.Problem, targetVariable);
            if (engine != null) {
              QueueJob(new Execution(dataSetEntry.Key, engine, targetVariable));
            }
          }
      }
    }

    private void QueueJob(Execution execution) {
      dispatchQueue.Add(execution);
    }

    public Execution GetNextJob() {
      if (dispatchQueue.Count == 0) FillDispatchQueue();
      if (dispatchQueue.Count > 0) {
        Execution next = dispatchQueue[0];
        dispatchQueue.RemoveAt(0);
        return next;
      } else
        return null;
    }

    internal void Start() {
      FillDispatchQueue();
    }

    private IEngine CreateEngine(Problem problem, int targetVariable) {
      switch (problem.LearningTask) {
        case LearningTask.Classification: return null;
        case LearningTask.Regression: {
            return CreateStandardGp(problem, targetVariable).Engine;
          }
        case LearningTask.TimeSeries: return null;
        case LearningTask.Clustering: return null;
        default: return null;
      }
    }

    private StandardGP CreateStandardGp(Problem problem, int targetVariable) {
      ProblemInjector probInjector = new ProblemInjector(problem);
      probInjector.TargetVariable = targetVariable;
      StandardGP sgp = new StandardGP();
      sgp.SetSeedRandomly = true;
      sgp.MaxGenerations = 100;
      sgp.PopulationSize = 1000;
      sgp.Elites = 1;
      sgp.ProblemInjector = probInjector;
      return sgp;
    }
  }
}
